/*
 * Copyright (c) 1996, 2003 VIA Networking, Inc. All rights reserved.
 *
 * This software may be redistributed and/or modified under 
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.  
 
 * File: device_module.c
 *
 * Purpose: driver entry for initial, open, close, tx and rx.
 *
 * Author: Lyndon Chen
 *
 * Date: Nov 10, 2003
 *
 * Functions:
 *   
 *   device_found1 - module initial (insmod) driver entry 
 *   device_remove1 - module remove entry 
 *   device_init_info - device structure resource allocation function
 *   device_free_info - device structure resource free function
 *   device_get_pci_info - get allocated pci io/mem resource
 *   device_print_info - print out resource
 *   device_open - allocate dma/descripter resource & initial mac/bbp function
 *   device_xmit - data tx function
 *   device_intr - interrupt handle function
 *   device_set_multi - set mac filter
 *   device_ioctl - ioctl entry
 *   device_close - shutdown mac/bbp & free dma/descripter resource
 *
 * Revision History:
 *      11-10-2003 Lyndon Chen: initial version
 *  
 */ 
 
#undef __NO_VERSION__

#if !defined(__DEVICE_MODULE_H)
#include "device_module.h"
#endif
#if !defined(__DEVICE_EXP_H)
#include "device_exp.h"
#endif


/*---------------------  Static Definitions -------------------------*/


MODULE_AUTHOR("VIA Networking Technologies, Inc., <lyndonchen@vntek.com.tw>");
MODULE_LICENSE("Proprietary");
MODULE_DESCRIPTION("VIA Networking Solomon-A/B/G Wireless LAN Adapter Driver");


/*---------------------  Static Classes  ----------------------------*/


/*---------------------  Static Variables  --------------------------*/

static int          device_nics = 0;
static PSDevice_info     pDevice_infos= NULL;


static CHIP_TBL chip_info_table[]= {
    { VT3253,       "VIA Networking Solomon-A/B/G Wireless LAN Adapter ",
        256, 1,     DEVICE_FLAGS_IP_ALIGN|DEVICE_FLAGS_TX_ALIGN },
    {0,NULL}
};


static struct pci_device_id device_id_table[] __devinitdata = {
{ 0x1106, 0x3253, PCI_ANY_ID, PCI_ANY_ID, 0, 0, (int)&chip_info_table[0]},
{ 0, }
};

static struct net_device *root_device_dev = NULL;

/*---------------------  Static Functions  --------------------------*/

static int device_found1(struct pci_dev *pcid, const struct pci_device_id *ent);
static BOOL device_init_info(struct pci_dev* pcid, PSDevice_info* ppDevice, PCHIP_TBL);
static void device_free_info(PSDevice_info pDevice);
static BOOL device_get_pci_info(PSDevice_info, struct pci_dev* pcid);
static void device_init_ref_dev(PSDevice_info pDevice);
static void device_print_info(PSDevice_info pDevice);
static int device_open(struct net_device *dev);
static int device_xmit(struct sk_buff *skb, struct net_device *dev);
static int device_xmit_80211(struct sk_buff *skb, struct net_device *dev);
static irqreturn_t device_intr(int irq, void *dev_instance, struct pt_regs *regs);
static void device_set_multi(struct net_device *dev);
static struct net_device_stats *device_get_stats(struct net_device *dev);
static int device_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);
static int device_close(struct net_device *dev);
static int device_hw_reset(PSDevice_info pDevice);
static int device_hw_init(PSDevice_info pDevice);
static void device_read_mac(PSDevice_info pDevice, U8 *dev_addr);


static char* get_chip_name(int chip_id) {
    int i;
    for (i=0;chip_info_table[i].name!=NULL;i++)
        if (chip_info_table[i].chip_id==chip_id)
            break;
    return chip_info_table[i].name; 
}

static void __devexit device_remove1(struct pci_dev *pcid)
{
    PSDevice_info pDevice=pci_get_drvdata(pcid);
    
    if (pDevice==NULL)
        return;
    device_free_info(pDevice);

}


static int 
device_found1(struct pci_dev *pcid, const struct pci_device_id *ent)
{
    static BOOL         bFirst=TRUE;
    struct net_device*  dev=NULL;
    PCHIP_TBL          pChip_info=(PCHIP_TBL) ent->driver_data;
    PSDevice_info       pDevice;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)    
    int         rc;
#endif        


    if (device_nics ++>= MAX_UINTS) {
        printk(KERN_NOTICE DEVICE_NAME ": already found %d NICs\n", device_nics);
        return -ENODEV;
    }
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    dev = alloc_etherdev(sizeof(DEVICE_MOD_INFO));
#else
    dev = init_etherdev(dev, sizeof(DEVICE_MOD_INFO));    
#endif        

    if (dev==NULL) {
        printk(KERN_ERR DEVICE_NAME ": allocate net device failed");
        return -ENODEV;
    }
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    /* Chain it all together */
    SET_MODULE_OWNER(dev);
    SET_NETDEV_DEV(dev, &pcid->dev);
#endif            
    
    if (bFirst) {
        printk(KERN_NOTICE "%s Ver. %s\n",DEVICE_FULL_DRV_NAM, DEVICE_VERSION);
        printk(KERN_NOTICE "Copyright (c) 2003 VIA Networking Technologies, Inc.\n");
        bFirst=FALSE;
    }

    if (!device_init_info(pcid, &pDevice, pChip_info)) {
        return -ENOMEM;
    }
    pDevice->dev=dev;
	pDevice->next_module = root_device_dev;
    pDevice->flags = pChip_info->flags ;
	root_device_dev = dev;
    dev->priv=pDevice;
    dev->irq=pcid->irq;
    
    if (pci_enable_device(pcid)) {
        device_free_info(pDevice);
        return -ENODEV;
    }

    if (device_get_pci_info(pDevice,pcid)==FALSE) {
        printk(KERN_ERR DEVICE_NAME ": Failed to find PCI device.\n");
        device_free_info(pDevice);
        return -ENODEV;
    }

    pDevice->port_offset = (DWORD)ioremap(pDevice->mem_addr & PCI_BASE_ADDRESS_MEM_MASK, pDevice->io_size);
    
    if(pDevice->port_offset == 0) {
       printk(KERN_ERR DEVICE_NAME ": Failed to IO remapping ..\n");
       device_free_info(pDevice);
        return -ENODEV;
    }    
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    rc = pci_request_regions(pcid, DEVICE_NAME);
    if (rc) {
        printk(KERN_ERR DEVICE_NAME ":__Failed to find PCI device\n");
        device_free_info(pDevice);
        return -ENODEV;        
    }
#else
    if (check_region(pDevice->io_addr, pDevice->io_size)) {
        printk(KERN_ERR DEVICE_NAME ": Failed to find PCI device\n");
        device_free_info(pDevice);
        return -ENODEV;                
    }
    request_region(pDevice->io_addr, pDevice->io_size, DEVICE_NAME);    
#endif    
    
    dev->base_addr = pDevice->io_addr;
    
#ifdef IO_MAP
    pDevice->port_offset = pDevice->io_addr;
#endif    

    if (device_hw_reset(pDevice) == 0) {        
        printk(KERN_ERR DEVICE_NAME ": Failed to access hardware..\n");
        device_free_info(pDevice);
        return -ENODEV;  
    } 
    
    if (device_hw_init(pDevice) == 0) {        
        printk(KERN_ERR DEVICE_NAME ": Failed to initial hardware..\n");        
        device_free_info(pDevice);
        return -ENODEV;  
    }
             
        
    device_read_mac(pDevice, dev->dev_addr);        
    device_init_ref_dev(pDevice);    
   
    dev->irq                = pcid->irq;     
    dev->open               = device_open;
    dev->hard_start_xmit    = device_xmit;
    dev->stop               = device_close;
    dev->get_stats          = device_get_stats;
    dev->set_multicast_list = device_set_multi;
    dev->do_ioctl           = device_ioctl;
    
    
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    rc = register_netdev(dev);
    if (rc)
    {
        printk(KERN_ERR DEVICE_NAME " Failed to register netdev\n");
        device_free_info(pDevice);
        return -ENODEV;          
    }
#endif        
    device_print_info(pDevice);
    pci_set_drvdata(pcid, pDevice);
    return 0;
}

static void device_print_info(PSDevice_info pDevice)
{
    struct net_device* dev=pDevice->dev;
  
    printk(KERN_INFO "%s: %s\n",dev->name, get_chip_name(pDevice->chip_id));
    printk(KERN_INFO "%s: MAC=%2.2X:%2.2X:%2.2X:%2.2X:%2.2X:%2.2X",
        dev->name,
        dev->dev_addr[0],dev->dev_addr[1],dev->dev_addr[2],
        dev->dev_addr[3],dev->dev_addr[4],dev->dev_addr[5]);
#ifdef IO_MAP
    printk(" IO=0x%lx  ",(ULONG) pDevice->io_addr);
    printk(" IRQ=%d \n", pDevice->dev->irq);
#else                
    printk(" IO=0x%lx Mem=0x%lx ",(ULONG) pDevice->io_addr,(ULONG) pDevice->port_offset);
    printk(" IRQ=%d \n", pDevice->dev->irq);
#endif    

}

static BOOL device_init_info(struct pci_dev* pcid, PSDevice_info* ppDevice,
            PCHIP_TBL pChip_info)
{
    PSDevice_info p;
        
    *ppDevice=kmalloc(sizeof(DEVICE_MOD_INFO),GFP_ATOMIC);
    
    if (*ppDevice==NULL)
        return FALSE;
        
    memset(*ppDevice,0,sizeof(DEVICE_MOD_INFO));
    
    if (pDevice_infos==NULL) {
        pDevice_infos=*ppDevice;      
    }
    else {
        for (p=pDevice_infos;p->next!=NULL;p=p->next)
            do {} while (0);    
        p->next=*ppDevice;
        (*ppDevice)->prev=p;
    }

    (*ppDevice)->pcid=pcid;
    (*ppDevice)->chip_id=pChip_info->chip_id;
    (*ppDevice)->io_size=pChip_info->io_size;
    (*ppDevice)->nTxQueues=pChip_info->nTxQueue;
    (*ppDevice)->multicast_limit=32;
    
    return TRUE;
}

static BOOL device_get_pci_info(PSDevice_info pDevice, struct pci_dev* pcid) {
    U16 pci_cmd;
    U8  b;
    UINT cis_addr; 

    
    pci_read_config_byte(pcid, PCI_REVISION_ID, &pDevice->byRevId);
    pci_read_config_word(pcid, PCI_SUBSYSTEM_ID, &pDevice->SubSystemID);
    pci_read_config_word(pcid, PCI_SUBSYSTEM_VENDOR_ID, &pDevice->SubVendorID);
    pci_read_config_word(pcid, PCI_COMMAND, (u16 *) & (pci_cmd));
    pci_set_master(pcid);
    pDevice->mem_addr=pci_resource_start(pcid,0);
    pDevice->io_addr=pci_resource_start(pcid,1);
    cis_addr =pci_resource_start(pcid,2);    
    
    pDevice->pcid=pcid;
    pci_read_config_byte(pcid, PCI_REG_COMMAND, &b);
    pci_write_config_byte(pcid, PCI_REG_COMMAND, (b|COMMAND_BUSM));
            
    return TRUE;                
}

static void device_init_ref_dev(PSDevice_info pDevice) {
    struct net_device*  dev=pDevice->dev;

    pDevice->ref_dev.last_rx = &(dev->last_rx);
    pDevice->ref_dev.trans_start = &(dev->trans_start);
    pDevice->ref_dev.flags = &(dev->flags);
    pDevice->ref_dev.mc_count = &(dev->mc_count);    
    pDevice->ref_dev.mc_list = dev->mc_list;        
    
    return ;
    
}

static void device_free_info(PSDevice_info pDevice) {
    PSDevice_info         ptr;
    struct net_device*  dev=pDevice->dev;

    ASSERT(pDevice);

    if (pDevice_infos==NULL)
        return;
    
    for (ptr=pDevice_infos;ptr && (ptr!=pDevice);ptr=ptr->next)
            do {} while (0);
    
    if (ptr==pDevice) {
        if (ptr==pDevice_infos)
            pDevice_infos=ptr->next;
        else 
            ptr->prev->next=ptr->next; 
    }
    else {
        printk(KERN_ERR DEVICE_NAME "device_info struct not found\n");
        return;
    }
    if (dev)
        unregister_netdev(dev);        
    
    if (pDevice->port_offset)
        iounmap((PVOID)pDevice->port_offset);

        
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    if (pDevice->pcid)   
        pci_release_regions(pDevice->pcid);
    if (dev)                
        free_netdev(dev);
#else
    if (pDevice->io_addr)
        release_region(pDevice->io_addr,pDevice->io_size);
    if (dev)        
        kfree(dev);
#endif        

    if (pDevice->pcid) {
        pci_set_drvdata(pDevice->pcid,NULL);
    }

    kfree(pDevice);
}

static int device_hw_reset(PSDevice_info pDevice) {

    return __device_hw_reset(pDevice);
}


static int device_hw_init(PSDevice_info pDevice) {
    
    return __device_hw_init(pDevice);
       
}

static void device_read_mac(PSDevice_info pDevice, U8 *dev_addr) {

     __device_read_mac(pDevice, dev_addr);
     
}

static int  device_open(struct net_device *dev) {
    PSDevice_info pDevice=(PSDevice_info) dev->priv;
    int i;    
    int result = 0;
   
    i = request_irq(pDevice->pcid->irq, &device_intr, SA_SHIRQ, dev->name, dev);
    if (i)
        return i;

    result = __device_open(pDevice);
    if (result == 0) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
        MOD_INC_USE_COUNT;
#endif  
    }
    
    
    return result;
}



static int device_close(struct net_device *dev) {
    PSDevice_info pDevice=(PSDevice_info) dev->priv;    

    __device_close(pDevice);
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_DEC_USE_COUNT;
#endif    
    return 0;     
}

static int device_xmit(struct sk_buff *skb, struct net_device *dev) {
    PSDevice_info pDevice=(PSDevice_info) dev->priv;
    
    __device_xmit(pDevice, skb);    

    return 0;
}

static int device_xmit_80211(struct sk_buff *skb, struct net_device *dev) {

    PSDevice_info pDevice=(PSDevice_info) dev->priv;

    __device_dma0_tx_80211(pDevice, skb);

    return 0;    
}    

    

static irqreturn_t device_intr(int irq, void *dev_instance, struct pt_regs *regs) {
    struct net_device *dev = dev_instance;
    PSDevice_info pDevice = dev->priv;
    int handled = 1;        

    handled = __device_intr(irq, pDevice, regs);
	
    return IRQ_RETVAL(handled);        

}

                    
static void device_set_multi(struct net_device *dev) {
    PSDevice_info pDevice=(PSDevice_info) dev->priv;

    __device_set_multi(pDevice);    

}

static struct net_device_stats *device_get_stats(struct net_device *dev) {
    PSDevice_info pDevice=(PSDevice_info) dev->priv;
    
    return __device_get_stats(pDevice);    

}



static int  device_ioctl(struct net_device *dev, struct ifreq *rq, int cmd) {
    
    PSDevice_info pDevice=(PSDevice_info) dev->priv;    
    
    return __device_ioctl(pDevice, rq, cmd);    
           
}



/*---------------------   export Functions  --------------------------*/



struct net_device *ref_init_apdev(struct net_device *dev){
    PSDevice_info pDevice=(PSDevice_info) dev->priv;

	pDevice->apdev = (struct net_device *)kmalloc(sizeof(struct net_device), GFP_KERNEL);
	if (pDevice->apdev == NULL)
		return NULL;
	memset(pDevice->apdev, 0, sizeof(struct net_device));
	pDevice->apdev->priv = pDevice;
	memcpy(pDevice->apdev->dev_addr, dev->dev_addr, ETH_ALEN);
	pDevice->apdev->type = ARPHRD_IEEE80211;
	pDevice->apdev->base_addr = dev->base_addr;
	pDevice->apdev->irq = dev->irq;
	pDevice->apdev->mem_start = dev->mem_start;
	pDevice->apdev->mem_end = dev->mem_end;
    pDevice->apdev->hard_start_xmit = device_xmit_80211;
  	sprintf(pDevice->apdev->name, "%sap", dev->name);
  	//printk("pDevice->apdev = %p, name = %s", pDevice->apdev, pDevice->apdev->name);

    return (pDevice->apdev);
}


struct net_device *ref_init_wpadev(struct net_device *dev){
    PSDevice_info pDevice=(PSDevice_info) dev->priv;

	pDevice->wpadev = (struct net_device *)kmalloc(sizeof(struct net_device), GFP_KERNEL);
	if (pDevice->wpadev == NULL)
		return NULL;
		
	memset(pDevice->wpadev, 0, sizeof(struct net_device));
	pDevice->wpadev->priv = pDevice;
	memcpy(pDevice->wpadev->dev_addr, dev->dev_addr, ETH_ALEN);
	pDevice->wpadev->type = ARPHRD_IEEE80211;
    pDevice->wpadev->base_addr = dev->base_addr;
	pDevice->wpadev->irq = dev->irq;
	pDevice->wpadev->mem_start = dev->mem_start;
	pDevice->wpadev->mem_end = dev->mem_end;
	sprintf(pDevice->wpadev->name, "%swpa", dev->name);		

    return (pDevice->wpadev);
}

void ref_kfree(void *buf) {
    
	kfree(buf);
}

void ref_dev_kfree_skb(struct sk_buff *skb) {
    
	dev_kfree_skb(skb);    
	
}   
   
void ref_dev_kfree_skb_irq(struct sk_buff *skb) {
    
	dev_kfree_skb_irq(skb);    
	
}      

U8  *ref_skb_put(struct sk_buff *skb, UINT len) {

    return skb_put(skb, len);
}    

void ref_skb_queue_head_init(struct sk_buff_head *list) {

    skb_queue_head_init(list);
}

void ref_skb_queue_tail(struct sk_buff_head *list, struct sk_buff *newsk) {
    
    skb_queue_tail(list, newsk);
}

int ref_skb_queue_empty(struct sk_buff_head *list) {
    
    return skb_queue_empty(list);
    
}    

void ref_init_timer(struct timer_list *timer) {

    init_timer(timer);
}    

void ref_del_timer(struct timer_list *timer) {
    
    del_timer(timer);
}    

void ref_add_timer(struct timer_list *timer) {
    
    add_timer(timer);
}

    
U8 *ref_kmalloc(int size, int flags) {
    
	U8 *pBuf;

	pBuf = kmalloc(size, flags);
	return pBuf;    
}    


struct sk_buff *ref_dev_alloc_skb(int size) {
    
       return dev_alloc_skb(size);
}

ref_sk_buff *ref_alloc_ref_skb(void) {

    ref_sk_buff     *ref_skb; 
    
    ref_skb = kmalloc(sizeof(ref_sk_buff), GFP_ATOMIC);
    
    return ref_skb;
}    

void ref_skb_remap(struct net_device *dev, ref_sk_buff *ref_skb, struct sk_buff *skb) {

    skb->dev = dev;
    ref_skb->dev = skb->dev;
    ref_skb->skb = skb;    
    ref_skb->data = skb->data;
    ref_skb->head = skb->head;    
    ref_skb->tail = skb->tail;
    ref_skb->len  = &(skb->len);
    ref_skb->protocol = &(skb->protocol);
    ref_skb->ip_summed = &(skb->ip_summed);
    ref_skb->pkt_type = &(skb->pkt_type);
    ref_skb->cb = &(skb->cb[0]);
    ref_skb->mac.raw = skb->mac.raw;
    
    return ;
}    

void ref_skb_add_offset(struct sk_buff* skb, int offset) {
    
    skb->data += offset;
    skb->tail += offset;
    return;
}    

void ref_skb_set_dev(struct net_device *dev, struct sk_buff* skb) {
    
    skb->dev = dev;
    return;
}    

void ref_set_rx_jiffies(struct net_device *dev) {
    
    dev->last_rx = jiffies;
    
    return;
}    

void ref_set_tx_jiffies(struct net_device *dev) {
    
    dev->trans_start = jiffies;
    
    return;
}   

void ref_netif_stop_queue(struct net_device *dev) {
    
    netif_stop_queue(dev);
}
    
void ref_netif_start_queue(struct net_device *dev) {
    
    netif_start_queue(dev);
}    
 
void ref_netif_wake_queue(struct net_device *dev) {
    
    netif_wake_queue(dev);
}

int ref_netif_queue_stopped(struct net_device *dev) {
    
    return netif_queue_stopped(dev);
}   


void ref_netif_rx(struct sk_buff *skb) {
    
    netif_rx(skb);
}

int ref_netif_running(struct net_device *dev) {
    
    return ref_netif_running(dev);   
}     
    
U16 ref_eth_type_trans(struct sk_buff *skb, struct net_device *dev){
    
    return eth_type_trans(skb, dev);   
} 
        
U32 ref_get_jiffies(void) {
    
    return jiffies;
} 
    
void ref_udelay(int cnt) {
    
    udelay(cnt);
}    

void ref_mdelay(int cnt) {
    
    mdelay(cnt);
}    


void *ref_pci_alloc_consistent(struct pci_dev *dev, 
                        size_t size, dma_addr_t *dma_handle) {
    
    return pci_alloc_consistent(dev, size, dma_handle);
}

void ref_pci_free_consistent(struct pci_dev *dev,
                     size_t size, void *vaddr, dma_addr_t dma_handle) {
                        
     pci_free_consistent(dev, size, vaddr, dma_handle);
                         
}                        

int ref_register_netdevice(struct net_device *dev) {
    
    return register_netdevice(dev);
    
}

int ref_register_netdev(struct net_device *dev) {
    
    return register_netdev(dev);    
}
    
int ref_unregister_netdevice(struct net_device *dev) {
    
    return unregister_netdevice(dev);    
}

void ref_unregister_netdev(struct net_device *dev) {
    
    unregister_netdev(dev);
}
    
    
int ref_HZ_tick(void) {
    
    return HZ;
}    
void ref_free_irq(unsigned int irq, struct net_device *dev) {
    
    free_irq(dev->irq, dev);
}    

int ref_copy_from_user(void *to, const void *from, int c) {
    
    return copy_from_user(to, from, c);
}    
    
int ref_copy_to_user(void *to, const void *from, int c) {
    
    return copy_to_user(to, from, c);
} 

void ref_spin_lock_init(spinlock_t *lock) {
    
     spin_lock_init(lock);
     
     return;
}     
    
void ref_pci_unmap_single(struct pci_dev *hwdev, dma_addr_t dma_addr, size_t size, int direction) {
    
     pci_unmap_single(hwdev, dma_addr, size, direction);
     
     return;
}     

dma_addr_t ref_pci_map_single(struct pci_dev *hwdev, void *ptr,	size_t size, int direction) {
    

     return pci_map_single(hwdev, ptr, size, direction);
     
}
    
/*------------------------------------------------------------------*/

MODULE_DEVICE_TABLE(pci, device_id_table);

static struct pci_driver device_driver = {
        name:       DEVICE_NAME,
        id_table:   device_id_table,
        probe:      device_found1,
        remove:     device_remove1,
        suspend:    NULL,
        resume:     NULL,
}; 

static int __init device_init_module(void)
{
    int ret;
    
    ret = pci_module_init(&device_driver);
    
    return ret;
}

static void __exit device_cleanup_module(void)
{

    pci_unregister_driver(&device_driver);

    
}
 
module_init(device_init_module);
module_exit(device_cleanup_module);

